/* CHD : Directory changing utility.  by  John H. DuBois III.      v1.1.
   Last rev 12/5/87.   Documentation in CHD.DOC */
#include <dir.h>
#include <dos.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
/* True if Path includes a drive (the second char is a colon) */
#define DriveSpeced(Path) (Path[0] != '\0' && Path[1] == ':')
#define UpCase(C) ('a' <= (C) && (C) <= 'z' ? (C) + 'A' - 'a' : (C))
#define FALSE (0)
#define TRUE (1)
typedef struct LN {	/* Node in list of paths generated from CDPATH var */
    char *String;
    struct LN *Next;
} ListNode;
typedef struct {	/* Linked list of paths */
   ListNode *First;
   ListNode *Current;
} StringList;
char *Insert(char *Target,char *Source,int Pos);
int TryPaths(StringList *CDPaths);
int MakeFullPaths(StringList *CDPaths,char *PathString,char *ArgPath);
int DNCSpeced(char *Path);
int PathNameSplit(char Path[],char Drive[],char Dir[]);
void AddDrive(char Path[]);
void AddWC(char *Path);
void AddString(StringList *List,char *S);
char *UsageMess =
"   CHD : Directory changing utility.  by  John H. DuBois III.   v1.1.\n"
"   Usage:  chd [<drive>:][<dir>][<control char>]";

main(argc,argv)
int argc;
char *argv[];
{
char Drive[MAXDRIVE],Direc[MAXDIR];  /* Drive and dir from command line arg */
int Compts;			/* Components found in command line arg */
/* Copy of com line arg created so drive and wildcards can be added to it */
char PathArg[MAXPATH];
static StringList CDPaths;	/* Linked list of paths from CDPATH var */
/* Whether arg ended in a control char, indicating that no dir name completion
   should be used */
int DirNameComp;

   /* Copy arg, or use empty string */
   strcpy(PathArg,argc == 1 ? "" : argv[1]);
   /* Check whether arg ended in a control char; if so remove the char and
      set dir name completion flag to true */
   DirNameComp = DNCSpeced(PathArg);
   /* Add drive to path if neccessay */
   AddDrive(PathArg);
   /* Split path into drive and direc components and determine whether any
      directory was given */
   Compts = PathNameSplit(PathArg,Drive,Direc);
   if ((Compts & ~DRIVE) == 0) {
      /* If no directory was given, print path for drive specified */
      if (getcurdir(*PathArg - 'A' + 1,Direc) == -1)
         fputs("Invalid drive.\n",stderr);
      else
         fprintf(stdout,"%s\\%s\n",Drive,Direc);
   }
   else {
      /* First attempt to change to a subdir of the current dir on the
         speficified drive */
      if (chdir(PathArg)) {
         /* If no success, try using env var "cdpath" */
         if (DirNameComp ^ (getenv("DNC") != NULL))
            /* If dir name completion is to be used, add wildcards to path */
            AddWC(PathArg);
         /* If path given is a relative path and CDPATH exists, try doing dir
            change using the paths in CDPATH */
         if (*Direc != '\\' && *Direc != '.' &&
          MakeFullPaths(&CDPaths,getenv("CDPATH"),PathArg) &&
          TryPaths(&CDPaths)) {
            /* If successful, print new full path */
            getcurdir(*PathArg - 'A' + 1,Direc);
            fprintf(stdout,"%s\\%s\n",Drive,Direc);
         }
         else {	/* Otherwise print error message */
            fputs("Invalid directory.\n",stderr);
            /* If first char of arg is not alphanumeric or dot, print usage. */
            if (!(isalnum(*(argv[1])) || (*(argv[1])) == '.'))
               fprintf(stderr,"%s\n",UsageMess);
         }
      }
   }
}

void AddWC(char *Path)
/* Adds wildcards to Path. */
{
   /* Used to split and reassemble path */
   char Drive[MAXDRIVE],Dir[MAXDIR],Name[MAXFILE],Ext[MAXEXT];
   int Len;

   fnsplit(Path,Drive,Dir,Name,Ext);	/* Get components of path */
   Len = strlen(Name);
   if (Len < 8) {	/* If name is less than 8 chars, add a '*' to it */
      Name[Len] = '*';
      Name[Len + 1] = '\0';
   }
   Len = strlen(Ext);
   if (0 < Len && Len < 4) {
      /* If extension is given and is less than 3 chars (plus the leading '.'),
         add a '*' to it */
      Ext[Len] = '*';
      Ext[Len + 1] = '\0';
   }
   fnmerge(Path,Drive,Dir,Name,Ext);
}

DNCSpeced(char *Path)
/* If Path ends with a control char, replaces the char with null char and
   returns true; else returns false */
{
   if (iscntrl(Path[strlen(Path) - 1])) {
      Path[strlen(Path) - 1] = '\0';
      return(TRUE);
   }
   else
      return(FALSE);
}

int MakeFullPaths(StringList *CDPaths,char *PathString,char *ArgPath)
/* Adds the drive and dir in ArgPath to the paths given in PathString and puts
   the new paths in a linked list (CDPaths). */
{
   char Drive[MAXDRIVE],	/* Drive speced in ArgPath */
   DriveLet,		/* Letter of the drive in ArgPath */
   Dir[MAXDIR],		/* Dir speced in ArgPath */
   *Path,		/* Current path from CDPATH env var */
   *FinalPath;		/* Path including drive */

   if (!PathString)	/* If PathString is null, return false */
      return(FALSE);
   PathNameSplit(ArgPath,Drive,Dir);	/* Get drive and dir */
   DriveLet = *Drive;			/* Get drive letter */
   Path = strtok(PathString,";");	/* Start tokenization */
   while(Path) {	/* While there are paths left in PathString */
      if (!DriveSpeced(Path) || UpCase(*Path) == DriveLet) {
         /* If no drive specified in path, or drive is specified and it
            matches the drive that chdir is begin done on, then the
            path is a candidate */
         /* Copy path to new string which includes room for tail dir name */
         FinalPath = (char *) calloc(MAXPATH,sizeof(char));
         if (DriveSpeced(Path))
            strcpy(FinalPath,Path);
         else {	/* If drive not speced use drive from PathString */
            strcpy(FinalPath,Drive);
            strcat(FinalPath,Path);
         }
         if (FinalPath[strlen(FinalPath) - 1] != '\\')
            /* If path did not already end in a backslash, add one */
            strcat(FinalPath,"\\");
         strcat(FinalPath,Dir);	/* Add dir from command line */
         AddString(CDPaths,FinalPath);
      }
      Path = strtok(NULL,";");	/* Get next path from PathString */
   }
   /* Return false (NULL) if CDPaths is empty (no paths, or only ones with
      different drive names, were found in PathString) */
   return((int) CDPaths -> First);
}

void AddDrive(char Path[])
/* If Path does not already include a drive, prefixes a drive spec (the letter
   of the current drive followed by a colon) to it.  If it does,
   the drive letter is changed to upper case if it was given in lower case. */
{
   char Drive[3];

   if (!DriveSpeced(Path)) {
      *Drive = getdisk() + 'A';
      Drive[1] = ':';
      Drive[2] = '\0';
      Insert(Path,Drive,0);
   }
   else {
      *Path = UpCase(*Path);
   }
}

char *Insert(char *Target,char *Source,int Pos)
/* Inserts string Source into string Target at position Pos. If the length of
   Target is less than Pos, Source is appended to Target. */
{
int SLen,TLen;	/* Length of source and target strings */

   SLen = strlen(Source);
   TLen = strlen(Target);
   if (TLen < Pos)
      strcat(Target,Source);
   else {
      memcpy(Target + Pos + SLen,Target + Pos,TLen - Pos + 1);	/* Make room
					in target string for source string */
      memcpy(Target + Pos,Source,SLen); /* Copy source string to targ string */
   }
   return(Target);
}

int PathNameSplit(char Path[],char Drive[],char Dir[])
/* Splits Path into its components and stores the drive part in Drive and the
   rest in Dir; returns a word indicating which parts were found. */
{
int Flags;
char File[MAXFILE],Ext[MAXEXT];

   Flags = fnsplit(Path,Drive,Dir,File,Ext);
   strcat(strcat(Dir,File),Ext);
   return(Flags);
}

int TryPaths(StringList *CDPaths)
/* Searches the paths in CDPaths for a valid path; if one is found changes
   to it. CDPaths can contain wildcards. */
{
   ListNode *Current;	/* Current node from CDPath being used */
   char *Path;		/* Path from Current */
   struct ffblk FFBlk;	/* Used by dir search function */
   int MoreMatches;	/* Whether there are more possible matches to Path */
   /* Used to split and reassemble a path */
   char NewPath[MAXPATH],Drive[MAXDRIVE],Dir[MAXDIR],Name[MAXFILE],Ext[MAXEXT];
   int ChDSuccess;	/* Whether dir change attempt was successful */

   for (Current = CDPaths -> First; Current; Current = Current -> Next) {
      /* For each path in CDPaths */
      Path = Current -> String;
      /* Get first matching dir. */
      MoreMatches = !findfirst(Path,&FFBlk,FA_DIREC);
      ChDSuccess = FALSE;
      while (MoreMatches && !ChDSuccess) {
         /* Do while there are matching "dirs" but the dir matched cannot
            be successfully changed to.  This apparently occurs when a non-
            dir is returned as a match even though only dirs are specified. */
         /* Replace name and ext given in Path with name and ext found */
         fnsplit(Path,Drive,Dir,Name,Ext);
         fnmerge(NewPath,Drive,Dir,FFBlk.ff_name,"");
         ChDSuccess = !chdir(NewPath);	/* Attempt to change to path */
         if (!ChDSuccess)	/* Get next matching "dir" */
            MoreMatches = !findnext(&FFBlk);
      }
      if (ChDSuccess)
         return(TRUE);
   }
   return(FALSE);
}

void AddString(StringList *List,char *S)
/* Adds string S to StringList */
{
   ListNode *Temp;

   Temp = (ListNode *) malloc(sizeof(ListNode));	/* Create a new node */
   Temp -> Next = NULL;
   Temp -> String = S;		/* Put S in it */
   if (List -> First)
      List -> Current -> Next = Temp;
   else
      List -> First = Temp;
   List -> Current = Temp;
}
